@@ -319,6 +319,7 @@ ddpgrow.default		<- function(y = NULL, subject, trt = NULL, time = NULL, n.rando
   subject		<- subject[o]
   map.subject		<- map.subject[o,]
   map.trt		<- map.trt[o,]
+  time			<- time[o] ## used for growth curve plotting
 
   ## capture number of fixed effects
   Nfixed		= ncol(X)
@@ -351,6 +352,7 @@ ddpgrow.default		<- function(y = NULL, subject, trt = NULL, time = NULL, n.rando
   summary.results$X			<- X
   summary.results$Z			<- Z
   summary.results$map.subject		<- map.subject
+  summary.results$time			<- time ## not used in accessor functions; just reporting back to user to let them know that sorted by subject
   summary.results$map.trt		<- map.trt
   summary.results$map.trtcov		<- map.trtcov
   summary.results$typet			<- typet
