@@ -284,6 +284,7 @@ dpgrow.default		<- function(y = NULL, subject, trt = NULL, time = NULL, n.random
   subject		<- subject[o]
   map.subject		<- map.subject[o,]
   map.trt		<- map.trt[o,]
+  time			<- time[o] ## used for growth curve plotting
 
   ## capture number of fixed effects
   Nfixed		= ncol(X)
@@ -309,6 +310,7 @@ dpgrow.default		<- function(y = NULL, subject, trt = NULL, time = NULL, n.random
   summary.results$X			<- X
   summary.results$Z			<- Z
   summary.results$map.subject		<- map.subject
+  summary.results$time			<- time ## not used in accessor functions; just reporting back to user to let them know that sorted by subject
   summary.results$map.trt		<- map.trt
   summary.results$model 		<- option
   summary.results$n.fix_degree		<- n.fix_degree
